package ca.yorku.cse.mack.graffitiexperiment;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Scanner;
import android.app.Activity;
import android.content.Context;
import android.content.Intent;
import android.content.SharedPreferences;
import android.graphics.Color;
import android.os.Bundle;
import android.text.Editable;
import android.text.TextWatcher;
import android.util.DisplayMetrics;
import android.util.Log;
import android.view.Gravity;
import android.view.KeyEvent;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.ViewGroup;
import android.view.ViewTreeObserver.OnGlobalLayoutListener;
import android.view.WindowManager;
import android.view.inputmethod.InputMethodManager;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.CheckBox;
import android.widget.EditText;
import android.widget.LinearLayout;
import android.widget.ScrollView;
import android.widget.Spinner;
import android.widget.TextView;
import android.widget.TextView.OnEditorActionListener;
import android.widget.Toast;

/**
 * Setup - a class that implements a setup dialog for experimental applications on Android.
 * <p>
 * 
 * A setup dialog is used to load parameters for an application and to present the parameters in a
 * UI that pops up before the application launches. The user is given the opportunity to change
 * setup parameters as appropriate for the current launch of the application. Changes to parameters
 * may be saved and used as default values for the next launch.
 * <p>
 * 
 * The motivation for the <code>Setup</code> class is to provide a single, consistent setup
 * mechanism for use across different experimental applications on Android. The class was first
 * developed for the Android application <code>GraffitiExperiment</code>. Consult for further
 * details.
 * 
 * What is needed:
 * <p>
 * 
 * <ul>
 * <li>A resource file named <code>spinnertextview.xml</code> specifying the spinner style for the
 * Spinner setup item. An example is found in GraffitiExperiment in
 * <code>/res/layout/spinnertextview.xml</code>.
 * <p>
 * <li>A resource file named <code>setupparameters.xml</code> containing an array of setup
 * parameters with the name <code>parameterarray</code>. The parameters are loaded in the main
 * application as follows:
 * 
 * <pre>
 *      String[] parameters = getResources().getStringArray(R.array.parameterarray);
 * </pre>
 * 
 * </ul>
 * 
 * @author Scott MacKenzie, 2014
 * 
 */
@SuppressWarnings("unused")
public class Setup extends Activity implements OnClickListener
{
	final static String MYDEBUG = "MYDEBUG"; // for Log.i messages
	final static String TITLE = "Setup";
	final static int BACKGROUND_COLOR = 0xff123456;
	final static int TEXT_SIZE = 18;
	final static int TEXT_SIZE_TITLE = 26;

	SharedPreferences sp;
	SharedPreferences.Editor spe;

	Button ok, save, exit;
	SetupItem[] setupItemArray;
	int width;
	
	ScrollView scrollView; // the root view

	/** Called when the activity is first created. */
	@Override
	public void onCreate(Bundle savedInstanceState)
	{
		super.onCreate(savedInstanceState);

		/*
		 * Initialize reference to shared preferences. NOTE: The values saved are the default
		 * values. The 1st time the app is run, there will be no shared preferences. To accommodate
		 * this, we use "?" as the default value, which is passed on as null when each SetupItem is
		 * created (see below).
		 */
		sp = this.getPreferences(MODE_PRIVATE);
		String[] defaultParameters = sp.getString("PARAMETERS", "?").split(",");

		DisplayMetrics dm = new DisplayMetrics();
		this.getWindowManager().getDefaultDisplay().getMetrics(dm);
		width = dm.widthPixels;

		// prevent soft keyboard from popping up when activity launches
		this.getWindow().setSoftInputMode(WindowManager.LayoutParams.SOFT_INPUT_STATE_ALWAYS_HIDDEN);

		scrollView = new ScrollView(this);
		scrollView.setBackgroundColor(BACKGROUND_COLOR);

		LinearLayout innerView = new LinearLayout(this);
		innerView.setOrientation(LinearLayout.VERTICAL);

		LinearLayout titleView = new LinearLayout(this);
		titleView.setGravity(Gravity.CENTER_HORIZONTAL);
		titleView.setPadding(0, 10, 0, 10);

		TextView title = new TextView(this);
		title.setText(TITLE);
		title.setTextSize(TEXT_SIZE_TITLE);
		titleView.addView(title);

		LinearLayout buttonView = new LinearLayout(this);
		buttonView.setGravity(Gravity.CENTER);
		buttonView.setPadding(0, 10, 0, 20);

		ok = new Button(this);
		ok.setText("OK");
		ok.setTextSize(TEXT_SIZE);
		ok.setOnClickListener(this);
		ok.setWidth(width / 3);

		save = new Button(this);
		save.setText("Save");
		save.setTextSize(TEXT_SIZE);
		save.setOnClickListener(this);
		save.setWidth(width / 3);

		exit = new Button(this);
		exit.setText("Exit");
		exit.setTextSize(TEXT_SIZE);
		exit.setOnClickListener(this);
		exit.setWidth(width / 3);

		// load parameters from resource and build setup items for UI
		String[] parameters = getResources().getStringArray(R.array.parameterarray);
		setupItemArray = new SetupItem[parameters.length];
		for (int i = 0; i < setupItemArray.length; ++i)
		{
			String[] resourceParameters = parameters[i].split("#");
			int type = Integer.parseInt(resourceParameters[0].trim());
			String label = resourceParameters[1].trim();
			String[] values = resourceParameters[2].split(",");
			String defaultValue = defaultParameters[0].equals("?") ? null : defaultParameters[i];
			setupItemArray[i] = new SetupItem(this, type, label, values, defaultValue);
		}

		buttonView.addView(ok);
		buttonView.addView(save);
		buttonView.addView(exit);

		innerView.addView(titleView);
		innerView.addView(buttonView);
		for (int i = 0; i < setupItemArray.length; ++i)
			innerView.addView(setupItemArray[i]);

		scrollView.addView(innerView);
		setContentView(scrollView);
	}

	@Override
	public void onClick(View v)
	{
		if (v == ok)
		{
			// send the chosen setup parameters to the main app as a comma-delimited string
			Bundle b = new Bundle();
			b.putString("PARAMETERS", this.getParameterString());

			// start experiment activity
			Intent i = new Intent(getApplicationContext(), GraffitiExperimentActivity.class);
			i.putExtras(b);
			startActivity(i);
			finish();

		} else if (v == save)
		{
			spe = sp.edit();
			spe = spe.putString("PARAMETERS", this.getParameterString());
			spe.commit();
			Toast.makeText(this, "Setup parameters saved!", Toast.LENGTH_SHORT).show();

		} else if (v == exit)
		{
			this.finish(); // terminate
		}
	}

	// get current parameter values as a comma-delimited string
	private String getParameterString()
	{
		String s = "";
		for (int i = 0; i < setupItemArray.length; ++i)
		{
			s += setupItemArray[i].getValue();
			if (i < setupItemArray.length - 1)
				s += ",";
		}
		return s;
	}

	// -------------------
	// S E T U P _ I T E M
	// -------------------

	class SetupItem extends LinearLayout
	{
		// the following avoids a "warning" with Java 1.5.0 complier
		private static final long serialVersionUID = 1L;

		final static int SPINNER = 100;
		final static int EDIT_TEXT = 200;
		final static int CHECK_BOX = 300;

		TextView comboBoxLabel;
		Spinner setupComboBox;

		TextView textFieldLabel;
		EditText setupEditText;

		CheckBox setupCheckBox;
		TextView checkBoxLabel;

		int type;
		String label;
		String[] values;
		String defaultValue;
		int preferredLeftWidth, preferredRightWidth;

		SetupItem(Context context, int typeArg, String labelArg, String[] valuesArg, String defaultValueArg)
		{
			super(context);
			type = typeArg;
			label = labelArg;
			values = valuesArg;
			defaultValue = defaultValueArg;
			// this.setPadding(2, 2, 2, 2);

			if (type == SPINNER)
			{
				comboBoxLabel = new TextView(context);
				comboBoxLabel.setWidth(width / 2);
				comboBoxLabel.setText(label);
				comboBoxLabel.setTextSize(TEXT_SIZE);
				comboBoxLabel.setGravity(Gravity.RIGHT);
				comboBoxLabel.setPadding(0, 0, 5, 0);
				this.addView(comboBoxLabel);

				setupComboBox = new Spinner(context);
				ArrayAdapter<CharSequence> adapter = new ArrayAdapter<CharSequence>(this.getContext(),
						R.layout.spinnertextview, values);
				setupComboBox.setAdapter(adapter);
				setupComboBox.setLayoutParams(new LayoutParams(width / 2 - 10, LayoutParams.WRAP_CONTENT));
				this.addView(setupComboBox);

				// set the default value
				if (defaultValue != null)
					setupComboBox.setSelection(adapter.getPosition(defaultValue));
				else
					setupComboBox.setSelection(0);
			}

			else if (type == EDIT_TEXT)
			{
				textFieldLabel = new TextView(context);
				textFieldLabel.setWidth(width / 2);
				textFieldLabel.setText(label);
				textFieldLabel.setTextSize(TEXT_SIZE);
				textFieldLabel.setGravity(Gravity.RIGHT);
				textFieldLabel.setPadding(0, 0, 5, 0);
				this.addView(textFieldLabel);

				setupEditText = new EditText(context);
				LinearLayout.LayoutParams llp = new LinearLayout.LayoutParams(width / 2 - 25, 78);
				llp.setMargins(8, 8, 0, 15);
				setupEditText.setLayoutParams(llp);
				setupEditText.setTextSize(TEXT_SIZE);
				setupEditText.setBackgroundColor(Color.LTGRAY);
				setupEditText.setPadding(0, 0, 0, 0);
				setupEditText.setGravity(Gravity.CENTER);
				setupEditText.setCursorVisible(false);
				// setupEditText.setOnFocusChangeListener(this);

				this.addView(setupEditText);

				// set the default value
				if (defaultValue != null)
					setupEditText.setText(defaultValue);
				else
					setupEditText.setText(values[0]);
			}

			else if (type == CHECK_BOX)
			{
				setupCheckBox = new CheckBox(context);
				this.addView(setupCheckBox);

				checkBoxLabel = new TextView(context);
				checkBoxLabel.setText(label);
				checkBoxLabel.setTextSize(TEXT_SIZE);
				checkBoxLabel.setGravity(Gravity.LEFT);
				checkBoxLabel.setPadding(5, 0, 0, 0);
				this.addView(checkBoxLabel);

				// set the default value, if specified
				if (defaultValue != null)
					setupCheckBox.setChecked(Boolean.valueOf(defaultValue));
				else
					setupCheckBox.setChecked(Boolean.valueOf(values[0]));
			}
		}

		// return the current value associated with this SetupItem
		public String getValue()
		{
			String s = "";
			if (type == SetupItem.CHECK_BOX)
				s = ("" + setupCheckBox.isChecked()).trim();
			else if (type == SetupItem.SPINNER)
				s = setupComboBox.getSelectedItem().toString().trim();
			else if (type == SetupItem.EDIT_TEXT)
				s = ("" + setupEditText.getText()).trim();
			return s;
		}
	}
}
