package ca.yorku.cse.mack.graffitiexperiment;

import java.io.*;
import java.util.*;
import android.graphics.Point;

/**
 * Unistroke - a class for performing handwriting recognition on gestures
 * <p>
 * 
 * A <i>Unistroke</i> is a set of <I>x-y</I> sample points created in a single
 * gesture with a stylus or finger on a digitizing surface. The sample points map out the shape of the gesture.  The gestures
 * represent characters, symbols, or commands. A gesture is also called a
 * <i>stroke</i>.
 * <p>
 * 
 * The idea for Unistrokes as a handwriting method for computers originated in 1993 in the paper <a
 * href=
 * "http://dl.acm.org/citation.cfm?id=164632.164677&coll=DL&dl=GUIDE&CFID=187414962&CFTOKEN=96829056"
 * >Touch Typing with a Stylus</a> by D. Goldberg and C. Richardson,  
 * There is also a 
 * <a href="http://www.google.ca/patents?hl=en&lr=&vid=USPAT5596656&id=RuECAAAAEBAJ&oi=fnd&dq=unistrokes&printsec=abstract#v=onepage&q=unistrokes&f=false">patent</a>
 * for Unistrokes, naming Goldberg as the inventor.  The patent issued in January 1997 and is assigned to Xerox Corporation.
 * <p>
 * 
 * The most widely-known
 * commercial example of Unistrokes is the 
 * <a href="http://en.wikipedia.org/wiki/Graffiti_%28Palm_OS%29">Graffiti</a> text entry method used
 * on the  
 * <a href="http://en.wikipedia.org/wiki/Palm_pilot">Palm Pilot</a>
 * personal digital assistant (PDA), first introduced in 1997.  The paper
 * <a href="http://www.yorku.ca/mack/GI97a.html">The Immediate Usability of Graffiti</a> provides
 * discussion and historical context for both Unistrokes and Graffiti.  The paper also presents an
 * empirical evaluation of Graffiti.
 * <p>
 * 
 * Applications using the <code>Unistroke</code> class generally need to
 * instantiate only one <code>Unistroke</code> object. Thereafter, all
 * processing, such as stroke recognition, loading custom dictionaries, and
 * switching between the built-in and custom dictionaries, is performed through
 * the fields and instance methods of the <code>Unistroke</code> class.
 * <p>
 * 
 * The <code>Unistroke</code> class includes three built-in dictionaries
 * (<code>GRAFFITI, UNISTROKES, DIGITS</code>) and support for five custom
 * dictionaries.
 * <p>
 * 
 * Recognition involves computing the <i>features</i> of an inputed stroke and
 * then comparing the features against those in the dictionary to find a match.
 * <p>
 * 
 * <h3>Dictionary Description</h3>
 * <p>
 * 
 * Each row in the dictionary contains thirteen (13) values delimited by spaces
 * or commas. As well, a dictionary may include comment lines beginning with 
 * "<code>#</code>" or blank lines. These are ignored.
 * <p>
 * 
 * The thirteen values consist of a recognized string for the stroke
 * (<code>symbol</code>) followed by twelve stroke features. The features consist
 * of four values representing the stroke's transition through quadrants of a
 * bounding box (<code>quadFirst, quadSecond, quadPenultimate, quadLast</code>),
 * four values representing the cumulative distance of the stroke along the
 * <I>x</I> and <I>y</I> axes (<code>kxmin, kxmax, kymin, kymax</code>), and
 * four values representing the starting and terminating directions of the
 * stroke along the <I>x</I> and <I>y</I> axes (<code>startx, starty, stopx,
 * stopy</code>).
 * <p>
 * 
 * In calculating the features in a stroke, the sample points are first
 * normalized to fit within a unit bounding box. The bounding box is divided
 * into four quadrants, "0" at the top right, "1" at the lower right, "2" at the
 * lower left, and "3" at the upper left:
 * <p>
 * 
 * <center> <a href="Unistroke-BoundingBox.jpg"><img src =
 * "Unistroke-BoundingBox.jpg" width="300"></a> </center>
 * <p>
 * 
 * Besides these four values representing a quadrant, a dictionary entry may
 * contain the value "4", implying "don't care". In this case, the recognizer
 * will not check the corresponding quadrant feature in the current stroke.
 * <p>
 * 
 * The dictionary entries are described in greater detail below.
 * <p>
 * 
 * <blockquote>
 * <dl>
 * <dt>
 * <code>symbol</code>
 * <dd>
 * the string the recognizer returns if the features in the current stroke match
 * the defined values in a particular row of the dictionary.
 * <p>
 * 
 * In most cases, the returned string is a single character representing the
 * inputed stroke, such as "a", "b", etc.
 * <p>
 * 
 * By convention, "=string" is used for strokes which are recognized, but for
 * which no particular character is assigned by the recognizer. For these
 * strokes, the application is expected to provide an appropriate
 * interpretation. For example,
 * 
 * <blockquote> <code>=SW</code> for a straight-line stroke in a south-west
 * direction </blockquote>
 * 
 * Since the recognizer returns a string, not a character, it is possible to
 * define shorthand strokes &mdash; single strokes that return a complete word
 * or phrase. These may be defined in or added to the home dictionary or in a
 * dedicated mode-shift dictionary. A mode-shift dictionary is a dictionary
 * invoked through a dedicated mode-shift stroke in the home dictionary (much
 * like the "symbol shift" stroke in Graffiti as implemented on the Palm Pilot).
 * <p>
 * 
 * The four quadrant features characterize the beginning and ending of a stroke
 * in terms of the expected quadrants. For example, the stroke for a Graffiti
 * "b" is expected to begin in quadrant 2, then transition into quadrant 3. The
 * stroke ends in quadrant 2, having transitioned from quadrant 1:
 * <p>
 * 
 * <center> <a href="Unistroke-Quadrant-Example.jpg"><img
 * src="Unistroke-Quadrant-Example.jpg" width="500"></a> </center>
 * <p>
 * 
 * The quadrant features are thus defined:
 * <p>
 * 
 * <dt>
 * <code>quadFirst</code>
 * <dd>
 * an integer representing the required first quadrant for the stroke.
 * <p>
 * 
 * <dt>
 * <code>quadSecond</code>
 * <dd>
 * an integer representing the required second quadrant for the stroke.
 * <p>
 * 
 * <dt>
 * <code>quadPenultimate</code>
 * <dd>
 * an integer representing the required penultimate (second last) quadrant for
 * the stroke.
 * <p>
 * 
 * <dt>
 * <code>quadLast</code>
 * <dd>
 * an integer representing the required last quadrant for the stroke.
 * </dl>
 * 
 * The four cumulative distance features are calculated in the recognizer by
 * summing the absolute distances between successive pairs of sample points
 * along the <I>x</I> and <I>y</I> axes.
 * <p>
 * 
 * The calculation is performed on the normalized sample points; that is, on the
 * sample points after they are scaled to fit within the unit bounding box
 * (width = 1, height = 1).
 * <p>
 * 
 * For example, a Graffiti "a", if entered perfectly, will have a cumulative
 * <I>x</I> distance of 1 unit and a cumulative <I>y</I> distance of 2 units:
 * <p>
 * 
 * <center> <a href="Unistroke-Alphabet-Graffiti-A.jpg"><img src =
 * "Unistroke-Alphabet-Graffiti-A.jpg" width="300"></a> </center>
 * <p>
 * 
 * Similarly, a Graffiti "z", if entered perfectly, will have a cumulative
 * <I>x</I> distance of 3 units and a cumulative <I>y</I> distance of 1 unit:
 * <p>
 * 
 * <center> <a href="Unistroke-Alphabet-Graffiti-Z.jpg"><img src =
 * "Unistroke-Alphabet-Graffiti-Z.jpg" width="300"></a> </center>
 * <p>
 * 
 * Obviously, users to do not enter their strokes perfectly. For a stroke to be
 * recognized, the calculated cumulative distance features must fall within the
 * minimum and maximum values (inclusive) in the stroke dictionary. These values
 * are specified as reals (e.g., 2.14).
 * <p>
 * 
 * Thus, the four cumulative distance values in the dictionary are defined as
 * follows.
 * <p>
 * <p>
 * 
 * <dl>
 * <dt>
 * <code>kxmin</code>
 * <dd>
 * the minimum cumulative <I>x</I> distance for the stroke
 * <p>
 * 
 * <dt>
 * <code>kxmax</code>
 * <dd>
 * the maximum cumulative <I>x</I> distance for the stroke
 * <p>
 * 
 * <dt>
 * <code>kymin</code>
 * <dd>
 * the minimum cumulative <I>y</I> distance for the stroke
 * <p>
 * 
 * <dt>
 * <code>kxmax</code>
 * <dd>
 * the maximum cumulative <I>y</I> distance for the stroke
 * </dl>
 * 
 * The last four dictionary values represent the starting and terminating
 * direction of the stroke. In the recognizer, these are computed on the first
 * and last 25% of the samples. The computed value is either 0 or 1, where "0"
 * implies motion to the left along the <I>x</I> axis or downward along the
 * <I>y</I> axis, and "1" implies motion to the right along the <I>x</I> axis or
 * upward along the <I>y</I> axis.
 * <p>
 * 
 * For example, the stroke for a Graffiti "a" should begin with motion
 * "to the right" along the <I>x</I> axis and motion "up" along the <I>y</I>
 * axis. It should terminate with motion "to the right" along the <I>x</I> axis
 * and motion "down" along the <I>y</I> axis:
 * <p>
 * 
 * <center> <a href="Unistroke-StartStopXY.jpg"><img src =
 * "Unistroke-StartStopXY.jpg" width="300"></a> </center>
 * <p>
 * 
 * Depending on the desired stroke shape, the corresponding dictionary entries
 * should be either 0 or 1. As well, "4" may appear in the dictionary as a
 * "don't care" condition. In this case, the recognizer does not check the
 * stroke's corresponding feature.
 * <p>
 * 
 * The four direction values are as follows.
 * <p>
 * 
 * <dl>
 * <dt>
 * <code>startx</code>
 * <dd>
 * an integer representing the required starting <i>x</i> direction for the
 * stroke
 * <p>
 * 
 * <dt>
 * <code>starty</code>
 * <dd>
 * an integer representing the required starting <i>y</i> direction for the
 * stroke
 * <p>
 * 
 * <dt>
 * <code>stopx</code>
 * <dd>
 * an integer representing the required terminating <i>x</i> direction for the
 * stroke
 * <p>
 * 
 * <dt>
 * <code>stopy</code>
 * <dd>
 * an integer representing the required terminating <i>y </i> direction for the
 * stroke
 * </dl>
 * </blockquote>
 * 
 * @author (c) Scott MacKenzie, 2001-2013
 * 
 */

public class Unistroke 
{
	public static final String UNRECOGNIZED_STROKE = "#";
	
	private final int TAP_THRESHHOLD = 5;		
	private final float ASPECT_RATIO = 0.2f;
	private int currentDictionary;

   // if the dictionary is used to find a match, this vector holds all the matches
   public static Vector<String> dictionaryMatches;
   
   /** A constant identifying a custom stroke dictionary.<p>
   *
   * The <code>Unistroke</code> class supports up to five custom
   * dictionaries.  They may be changed
   * on-the-fly using the <code>setDictionary()</code> method with this
   * field as an argument.
   * <p>
   * @see #setDictionary
   * @see #loadDictionary
   */
   public static final int CUSTOM1 = 1;

   /** A constant identifying a custom stroke dictionary.<p>
   *
   * The <code>Unistroke</code> class supports up to five custom
   * dictionaries.  They may be changed
   * on-the-fly using the <code>setDictionary()</code> method with this
   * field as an argument.
   * <p>
   * @see #setDictionary
   * @see #loadDictionary
   */
   public static final int CUSTOM2 = 2;

   /** A constant identifying a custom stroke dictionary.<p>
   *
   * The <code>Unistroke</code> class supports up to five custom
   * dictionaries.  They may be changed
   * on-the-fly using the <code>setDictionary()</code> method with this
   * field as an argument.
   * <p>
   * @see #setDictionary
   * @see #loadDictionary
   */
   public static final int CUSTOM3 = 3;

   /** A constant identifying a custom stroke dictionary.<p>
   *
   * The <code>Unistroke</code> class supports up to five custom
   * dictionaries.  They may be changed
   * on-the-fly using the <code>setDictionary()</code> method with this
   * field as an argument.
   * <p>
   * @see #setDictionary
   * @see #loadDictionary
   */
   public static final int CUSTOM4 = 4;

   /** A constant identifying a custom stroke dictionary.<p>
   *
   * The <code>Unistroke</code> class supports up to five custom
   * dictionaries.  They may be changed
   * on-the-fly using the <code>setDictionary()</code> method with this
   * field as an argument.
   * <p>
   * @see #setDictionary
   * @see #loadDictionary
   */
   public static final int CUSTOM5 = 5;

   /** A constant identifying the built-in Graffiti
   * stroke dictionary.<p>
   *
   * Use as an argument to the <code>setDictionary()</code> method.<p>
   *
   * The Graffiti alphabet is shown below:<p>
   *
   * <center>
   * <a href="Unistroke-Alphabet.Graffiti.jpg"><img src = "Unistroke-Alphabet-Graffiti.jpg" width="600"></a>
   * </center>
   * <p>
   * 
   * The stroke shapes above do not exactly match those in the original Graffiti alphabet.  Slight modifications
   * have been introduced to improve recognition.
   * <p>
   * 
   * A few entries above (e.g., Caps) include two strokes.  The interpretation is for a mode shift, which must
   * be implemented and managed by the application using the <code>Unistroke</code> class.
   * <p>
   *  
   * @see #setDictionary
   */
   public static final int GRAFFITI = 6;

   /** A constant identifying the built-in Unistrokes
   * stroke dictionary.<p>
   *
   * Use as an argument to the <code>setDictionary()</code> method.<p>
   *
   *
   * The Unistrokes alphabet is shown below:<p>
   *
   * <center>
   * <a href="Unistroke-Alphabet-Unistrokes.jpg"><img src = "Unistroke-Alphabet-Unistrokes.jpg" width="600"></a>
   * </center>
   * <p>
   * @see #setDictionary
   */
   public static final int UNISTROKES = 7;

   /** A constant identifying the built-in stroke dictionary for digits.<p>
   *
   * Use as an argument to the <code>setDictionary()</code> method.<p>
   *
   * The digit strokes are is shown below:<p>
   *
   * <center>
   * <a href="Unistroke-Alphabet-Digits.jpg"><img src = "Unistroke-Alphabet-Digits.jpg" width="600"></a>
   * </center>
   * <p>
   * @see #setDictionary
   */
   public static final int DIGITS = 8;

   /** Load a custom dictionary from a disk file.
   * After loading, the custom dictionary becomes the currently active
   * dictionary.<p>
   *
   * @param s the name of the disk file containing the stroke definitions.
   * @param n an integer representing the dictionary (e.g.,
   * <code>Unistroke.CUSTOM1</code>).  This argument is used later with 
   * <code>setDictionary()</code> to switch from one
   * dictionary to another.
   * @return <code>true</code> if the load was successful, <code>false</code>
   * otherwise.
   */
   public boolean loadDictionary(String s, int n) 
   {
      // attempt to open disk file for input
      BufferedReader inFile;
      try
      {
         inFile = new BufferedReader(new FileReader(s));
      }
      catch (FileNotFoundException e)
      {
         System.out.println("Can't find dictionary file: " + s);
         return false;
      }

      // we'll use a vector to collect the dictionary items
      Vector<StrokeDef> v = new Vector<StrokeDef>();

      String line;
      boolean moreData = true;
      while (moreData)
      {
         // try/catch for readLine()
         try
         {
            line = inFile.readLine();
         }
         catch (IOException e)
         {
            System.out.println("IOException error reading dictionary file");
            return false;
         }

         if (line == null)
         {
            moreData = false;
            continue;
         }
       
         // ignore blank lines
         if (line.length() == 0)
            continue;

         // ignore comments lines ("#" in 1st position)
         if (line.length() > 0 && line.substring(0, 1).equals("#"))
            continue;

         StringTokenizer st = new StringTokenizer(line, ", ");

         // ignore blank lines with spaces
         if (st.countTokens() == 0)
            continue;

         if (st.countTokens() != 13)
         {
            System.out.println("Data format error in file: " + s);
            System.out.println("... data: " + line);
            System.out.println("... token count: " + st.countTokens());
            System.out.println("... line length: " + line.length());
            return false;
         }

         StrokeDef sd = new StrokeDef(
            st.nextToken(),
            Integer.parseInt(st.nextToken()),
            Integer.parseInt(st.nextToken()),
            Integer.parseInt(st.nextToken()),
            Integer.parseInt(st.nextToken()),
            Double.parseDouble(st.nextToken()),
            Double.parseDouble(st.nextToken()),
            Double.parseDouble(st.nextToken()),
            Double.parseDouble(st.nextToken()),
            Integer.parseInt(st.nextToken()),
            Integer.parseInt(st.nextToken()),
            Integer.parseInt(st.nextToken()),
            Integer.parseInt(st.nextToken())    );
        v.add(sd);
      }

      // create a StrokeDef array of just the size needed
      custom1 = new StrokeDef[v.size()];

      // copy vector contents into array
      v.copyInto(custom1);

      // make this the active dictionary
      setDictionary(n);

      return true;
   }

   // digits stroke dictionary
   private static StrokeDef[] dsd = 
   {
	  new StrokeDef("0",  3,  2,  1,  0,  1.5, 2.5, 1.5,  2.5,   0,   0,   0,   1 ),
      new StrokeDef("0",  3,  2,  0,  3,  1.5, 2.5, 1.5,  2.5,   0,   4,   0,   4 ),
      new StrokeDef("0",  0,  3,  1,  0,  1.5, 2.5, 1.5,  2.5,   0,   4,   0,   4 ),
      new StrokeDef("0",  0,  3,  0,  3,  1.5, 2.5, 1.5,  2.5,   0,   4,   0,   4 ),
      new StrokeDef("0",  0,  1,  3,  0,  1.5, 2.5, 1.5,  2.5,   1,   4,   1,   4 ),
      new StrokeDef("0",  0,  1,  2,  3,  1.5, 2.5, 1.5,  2.5,   1,   0,   1,   1 ),
      new StrokeDef("0",  3,  0,  3,  0,  1.5, 2.5, 1.5,  2.5,   1,   0,   1,   1 ),
      new StrokeDef("0",  3,  0,  2,  3,  1.5, 2.5, 1.5,  2.5,   1,   0,   1,   1 ),
      // "1" is a straight-line stroke down (SOUTH).  See toCharacter
      new StrokeDef("2",  3,  0,  2,  1,  2.0, 4.0, 0.9,  1.5,   1,   4,   1,   4 ),
      new StrokeDef("2",  3,  0,  2,  1,  2.0, 4.0, 0.9,  1.5,   1,   4,   1,   4 ),
      new StrokeDef("3",  3,  0,  1,  2,  2.0, 4.5, 0.9,  1.5,   1,   4,   0,   4 ),
      new StrokeDef("4",  3,  2,  2,  1,  0.9, 1.5, 0.9,  1.5,   4,   0,   1,   4 ),
      new StrokeDef("4",  0,  3,  2,  1,  0.9, 2.0, 0.9,  1.5,   0,   0,   1,   4 ),
      new StrokeDef("5",  0,  3,  1,  2,  2.0, 3.5, 0.9, 1.30,   0,   4,   0,   4 ),
      new StrokeDef("5",  3,  0,  1,  2,  2.0, 3.5, 0.9, 1.30,   0,   4,   0,   4 ),
      new StrokeDef("6",  0,  3,  4,  2,  1.5, 3.0, 1.31, 2.5,   0,   4,   0,   4 ),
      new StrokeDef("6",  3,  2,  4,  2,  1.5, 2.5, 1.31, 2.5,   0,   4,   0,   4 ),
      new StrokeDef("7",  3,  0,  0,  1,  0.9, 1.5, 0.9,  1.5,   1,   4,   4,   0 ),
      new StrokeDef("7",  3,  0,  1,  2,  0.9, 2.0, 0.9,  1.5,   1,   4,   4,   0 ),
      new StrokeDef("7",  3,  0,  3,  2,  0.9, 2.0, 0.9,  1.5,   1,   4,   4,   0 ),
      new StrokeDef("8",  0,  4,  4,  0,  2.5, 4.0, 1.5,  3.5,   4,   4,   4,   1 ),
      new StrokeDef("8",  0,  4,  4,  3,  2.5, 4.0, 1.5,  3.5,   4,   4,   4,   1 ),
      new StrokeDef("8",  3,  4,  4,  0,  2.5, 4.0, 1.5,  3.5,   4,   4,   4,   1 ),
      new StrokeDef("8",  3,  4,  4,  3,  2.5, 4.0, 1.5,  3.5,   4,   4,   4,   1 ),
      new StrokeDef("9",  0,  3,  0,  1,  1.5, 2.5, 1.3,  2.5,   0,   4,   4,   0 )
   };

   // =========================================================================================
   // NOTE: The ten stroke features, in order, are...
   // quadFirst        - first quadrant (0 = top-right, 1 = bottom-right, 2 = bottom-left, 3 = top-left)
   // quadSecond       - second quadrant
   // quadsPenultimate - second-last quadrant
   // quadLast         - last quadrant
   // kx               - cumulative x motion (1 = once across in the x direction)
   // ky               - cumulative y motion (1 = once across in the y direction)
   // startx           - starting x direction (0 = left, 1 = right, 4 = don't care)
   // starty           - starting y direction (0 = down, 1 = up,  4 = don't care)
   // stopx            - stopping x direction
   // stopy            - stopping y direction
   // ==========================================================================================

   // ==========================
   // graffiti stroke dictionary
   // ==========================
   
   private static StrokeDef[] gsd =
   {
	   // Jan. 28, 2013 -- all straight-line strokes removed from the dictionary.  
	   // To improve recognition, straight-line strokes are handled by pre-processing
	   // the sample points.  See toCharacter

      new StrokeDef("a",  2,  3,  0,  1,  0.9, 1.2, 1.5,  2.0,   4,   1,   4,   0 ),
      new StrokeDef("b",  3,  2,  1,  2,  2.3, 4.0, 2.5,  3.5,   4,   0,   0,   4 ), // "B" starting at the top-left
      new StrokeDef("b",  2,  3,  1,  2,  2.3, 4.0, 1.5,  2.5,   4,   1,   0,   4 ), // "B" starting at the bottom-left
      new StrokeDef("c",  0,  3,  2,  1,  1.5, 2.0, 0.9,  2.0,   0,   4,   1,   4 ),
      new StrokeDef("d",  3,  2,  1,  2,  1.5, 2.3, 2.0,  3.0,   4,   0,   0,   4 ), // "D" starting at the top-left
      new StrokeDef("d",  2,  3,  1,  2,  1.5, 2.3, 1.5,  2.5,   4,   1,   0,   4 ), // "D" starting at the bottom-left
      new StrokeDef("e",  0,  3,  2,  1,  2.7, 4.0, 0.9,  2.0,   0,   4,   1,   4 ),
      new StrokeDef("f",  0,  3,  3,  2,  0.9, 1.3, 0.9, 1.25,   0,   4,   4,   0 ),
      new StrokeDef("g",  0,  3,  1,  2,  1.5, 3.5, 1.3,  2.5,   0,   4,   0,   4 ), // lowercase "g"
      new StrokeDef("g",  0,  3,  2,  1,  1.8, 2.7, 1.2,  2.2,   0,   4,   1,   4 ), // uppercase "G"
      new StrokeDef("h",  3,  2,  0,  1,  0.9, 1.5, 1.2,  2.5,   4,   0,   4,   0 ),
      new StrokeDef("h",  3,  2,  2,  1,  0.9, 1.5, 1.2,  2.5,   4,   0,   4,   0 ),
      new StrokeDef("h",  3,  2,  3,  1,  0.9, 1.5, 0.9,  2.5,   4,   0,   4,   0 ),
      // "i" is a straight-line stroke down (SOUTH).  See toCharacter
      new StrokeDef("j",  0,  1,  1,  2,  0.9, 1.4, 0.9, 1.25,   4,   0,   0,   4 ),
      new StrokeDef("k",  0,  4,  4,  1,  1.5, 2.5, 1.5,  2.5,   0,   0,   1,   0 ),
      new StrokeDef("l",  3,  2,  2,  1,  0.9, 1.3, 0.9, 1.25,   4,   0,   1,   4 ),
      new StrokeDef("m",  2,  3,  0,  1,  0.9, 1.5, 2.0,  4.0,   4,   1,   4,   0 ),
      new StrokeDef("m",  3,  2,  0,  1,  0.9, 1.5, 3.0,  4.5,   4,   0,   4,   0 ),
      new StrokeDef("n",  2,  3,  1,  0,  0.9, 1.5, 2.3,  4.0,   4,   1,   4,   1 ),
      new StrokeDef("o",  3,  2,  1,  0,  1.5, 2.5, 1.5,  2.5,   0,   0,   0,   1 ),
      new StrokeDef("o",  3,  2,  0,  3,  1.5, 2.5, 1.5,  2.5,   0,   4,   0,   4 ),
      new StrokeDef("o",  0,  3,  1,  0,  1.5, 2.5, 1.5,  2.5,   0,   4,   0,   4 ),
      new StrokeDef("o",  0,  3,  0,  3,  1.5, 2.5, 1.5,  2.5,   0,   4,   0,   4 ),
      new StrokeDef("o",  0,  1,  3,  0,  1.5, 2.5, 1.5,  2.5,   1,   4,   1,   4 ),
      new StrokeDef("o",  0,  1,  2,  3,  1.5, 2.5, 1.5,  2.5,   1,   0,   1,   1 ),
      new StrokeDef("o",  3,  0,  3,  0,  1.5, 2.5, 1.5,  2.5,   1,   0,   1,   1 ),
      new StrokeDef("o",  3,  0,  2,  3,  1.5, 2.5, 1.5,  2.5,   1,   4,   4,   1 ),
      new StrokeDef("p",  3,  2,  2,  3,  1.6, 3.0, 2.0,  3.5,   4,   0,   0,   4 ),
      new StrokeDef("p",  3,  2,  0,  3,  1.6, 3.0, 2.0,  3.5,   4,   0,   0,   4 ),
      new StrokeDef("p",  3,  2,  1,  3,  1.6, 3.0, 2.0,  3.5,   4,   0,   0,   4 ),
      new StrokeDef("p",  2,  3,  0,  3,  1.6, 3.0, 1.2,  3.5,   4,   1,   0,   4 ),
      new StrokeDef("q",  0,  3,  1,  0,  1.5, 2.5, 1.8,  3.0,   0,   0,   1,   4 ),
      new StrokeDef("q",  0,  3,  3,  0,  1.5, 2.5, 1.8,  3.0,   0,   0,   1,   4 ),
      new StrokeDef("q",  3,  2,  1,  0,  1.5, 2.5, 1.8,  3.0,   0,   0,   1,   4 ),
      new StrokeDef("q",  3,  2,  3,  0,  1.5, 2.5, 1.8,  3.0,   0,   0,   1,   4 ),      
      new StrokeDef("q",  0,  3,  4,  1,  2.2, 4.0, 2.5,  5.0,   0,   4,   1,   0 ),
      new StrokeDef("q",  3,  2,  4,  1,  2.2, 4.0, 2.5,  5.0,   0,   4,   1,   0 ),     
      new StrokeDef("r",  2,  3,  4,  1,  2.0, 3.0, 1.6,  3.0,   4,   1,   1,   0 ),
      new StrokeDef("s",  0,  3,  1,  2,  2.0, 3.5, 0.9,  2.0,   0,   4,   0,   4 ),
      new StrokeDef("t",  3,  0,  0,  1,  0.9, 1.2, 0.9, 1.25,   1,   4,   4,   0 ),
      new StrokeDef("u",  3,  2,  1,  0,  0.9, 2.5, 0.9,  2.0,   4,   0,   4,   1 ),
      new StrokeDef("v",  3,  4,  4,  0,  0.9, 1.2, 1.5,  2.0,   1,   0,   1,   0 ),
      new StrokeDef("v",  0,  1,  2,  3,  0.9, 1.4, 1.5,  2.0,   4,   0,   4,   1 ),
      new StrokeDef("w",  3,  2,  1,  0,  0.9, 1.5, 2.0,  4.0,   4,   0,   4,   1 ),
      new StrokeDef("x",  3,  4,  4,  2,  1.5, 2.5, 1.5,  3.5,   1,   0,   0,   0 ),
      new StrokeDef("y",  3,  4,  4,  2,  0.9, 2.0, 1.5,  2.5,   1,   0,   0,   4 ),
      new StrokeDef("y",  3,  4,  4,  0,  1.5, 2.8, 1.5,  2.9,   1,   0,   1,   1 ),
      new StrokeDef("y",  3,  4,  4,  1,  1.4, 3.0, 1.5,  2.5,   1,   0,   1,   4 ),
      new StrokeDef("z",  3,  0,  2,  1,  2.2, 3.5, 0.9,  2.0,   1,   4,   1,   4 )
   };

   // ============================
   // unistrokes stroke dictionary
   // ============================
   
   private static StrokeDef[] usd =
   {
      new StrokeDef("a", 1,  4,  4,  0,  0.0, 0.0, .98,  1.0,   4,   4,   4,   4 ),
      new StrokeDef("a", 2,  4,  4,  0,  0.0, 0.0, .98,  1.0,   4,   4,   4,   4 ),
      new StrokeDef("a", 1,  4,  4,  3,  0.0, 0.0, .98,  1.0,   4,   4,   4,   4 ),
      new StrokeDef("a", 2,  4,  4,  3,  0.0, 0.0, .98,  1.0,   4,   4,   4,   4 ),                                                           
      new StrokeDef("i", 0,  4,  4,  1,  0.0, 0.0, .98,  1.0,   4,   4,   4,   4 ),
      new StrokeDef("i", 3,  4,  4,  1,  0.0, 0.0, .98,  1.0,   4,   4,   4,   4 ),
      new StrokeDef("i", 0,  4,  4,  2,  0.0, 0.0, .98,  1.0,   4,   4,   4,   4 ),
      new StrokeDef("i", 3,  4,  4,  2,  0.0, 0.0, .98,  1.0,   4,   4,   4,   4 ),
      new StrokeDef("t", 3,  4,  4,  0,  .98, 1.0, 0.0,  0.0,   4,   4,   4,   4 ),
      new StrokeDef("t", 2,  4,  4,  0,  .98, 1.0, 0.0,  0.0,   4,   4,   4,   4 ),
      new StrokeDef("t", 3,  4,  4,  1,  .98, 1.0, 0.0,  0.0,   4,   4,   4,   4 ),
      new StrokeDef("t", 2,  4,  4,  1,  .98, 1.0, 0.0,  0.0,   4,   4,   4,   4 ),
      new StrokeDef("e", 0,  4,  4,  3,  .98, 1.0, 0.0,  0.0,   4,   4,   4,   4 ),
      new StrokeDef("e", 1,  4,  4,  3,  .98, 1.0, 0.0,  0.0,   4,   4,   4,   4 ),
      new StrokeDef("e", 0,  4,  4,  2,  .98, 1.0, 0.0,  0.0,   4,   4,   4,   4 ),
      new StrokeDef("e", 1,  4,  4,  2,  .98, 1.0, 0.0,  0.0,   4,   4,   4,   4 ),
      new StrokeDef("k", 2,  4,  4,  0, .99, 1.0, .99,  1.0,   4,   4,   4,   4 ),
      new StrokeDef("r", 3,  4,  4,  1, .99, 1.0, .99,  1.0,   4,   4,   4,   4 ),
      new StrokeDef("y", 0,  4,  4,  2, .99, 1.0, .99,  1.0,   4,   4,   4,   4 ),
      new StrokeDef("b", 3, 0, 1, 2, 1.5, 2.5, 0.9, 1.5, 1, 0, 0, 0),
      new StrokeDef("c", 1, 2, 3, 0, 1.5, 2.5, 0.9, 1.5, 0, 1, 1, 1),
      new StrokeDef("d", 0, 3, 2, 1, 1.5, 2.5, 0.9, 1.5, 0, 0, 1, 0),
      new StrokeDef("f", 0, 3, 3, 2, 0.9, 1.5, 0.9, 1.5, 0, 4, 4, 0),
      new StrokeDef("g", 2, 1, 1, 0, 0.9, 1.5, 0.9, 1.5, 1, 4, 4, 1),
      new StrokeDef("h", 3, 0, 0, 1, 0.9, 1.5, 0.9, 1.5, 1, 4, 4, 0),
      new StrokeDef("j", 0, 1, 1, 2, 0.9, 1.5, 0.9, 1.5, 4, 0, 0, 4),
      new StrokeDef("l", 3, 2, 2, 1, 0.9, 1.5, 0.9, 1.5, 4, 0, 1, 4),
      new StrokeDef("m", 1, 0, 3, 2, 0.9, 1.5, 1.5, 2.5, 0, 1, 0, 0),
      new StrokeDef("n", 2, 3, 0, 1, 0.9, 1.5, 1.5, 2.5, 1, 1, 1, 0),
      new StrokeDef("o", 0, 4, 4, 3, 1.5, 2.5, 1.5, 2.5, 0, 0, 0, 1),
      new StrokeDef("p", 3, 4, 4, 2, 1.5, 2.5, 1.5, 2.5, 1, 0, 0, 0),
      new StrokeDef("q", 0, 4, 4, 1, 1.5, 2.5, 1.5, 2.5, 0, 0, 1, 0),
      new StrokeDef("r", 3, 4, 4, 1, 0.9, 1.5, 0.9, 1.5, 1, 0, 1, 0),
      new StrokeDef("s", 0, 3, 1, 2, 2.0, 3.0, 0.9, 1.5, 0, 4, 0, 4),
      new StrokeDef("u", 0, 1, 2, 3, 0.9, 1.5, 1.5, 2.5, 0, 0, 0, 1),
      new StrokeDef("v", 3, 2, 1, 0, 0.9, 1.5, 1.5, 2.0, 1, 0, 1, 1),
      new StrokeDef("w", 3, 2, 0, 1, 0.9, 1.5, 2.0, 3.0, 4, 0, 4, 0),
      new StrokeDef("x", 3, 4, 4, 0, 1.5, 2.5, 1.5, 2.5, 1, 0, 1, 1),
      new StrokeDef("y", 2, 4, 4, 0, 0.9, 1.5, 0.9, 1.5, 1, 1, 1, 1),
      new StrokeDef("z", 3, 0, 2, 1, 2.0, 3.0, 0.9, 1.5, 1, 4, 1, 4),
      new StrokeDef("=CR", 2, 1, 0, 3, 1.5, 2.5, 0.9, 1.5, 1, 1, 0, 1)
   };

   /** Provides access to the currently active stroke dictionary.<p>
   *
   * A variety of debugging and advanced programming
   * services are available through this public variable.
   *
   * Given a <code>Unistroke</code> object <code>u</code>, the
   * number of entries in the currently active dictionary is
   *
   * <pre>
   *     u.activeDictionary.length
   * </pre>
   *
   * Each entry in the currently active dictionary may be
   * retrieved.  For example
   *
   * <pre>
   *     for (int i = 0; i < u.activeDictionary.length; ++i)
   *        System.out.println(u.activeDictionary[i]);
   * </pre>
   *
   * outputs the entire dictionary to the console.<p>
   *
   * Each row in the dictionary is returned as a string containing
   * thirteen comma-delimited values.  For example, the entry
   * for the letter "a" in the built-in Graffiti dictionary appears
   * as follows:<p>
   *
   * <pre>
   *     a, 2, 3, 0, 1, 0.9, 1.2, 1.5, 2.0, 4, 1, 4, 0
   * </pre>
   *
   * where
   *
   * <pre>
   *     symbol = a
   *     quadFirst  = 2
   *     quadSecond = 3
   *     quadsPenultimate = 0
   *     quadLast = 1
   *     kxmin = 0.9
   *     kxmax = 1.2
   *     kymin = 1.5
   *     kymax = 2.0
   *     startx = 4
   *     starty = 1
   *     stopx = 4
   *     stopy = 0
   * </pre>
   *
   * See the description for the <code>loadDictionary()</code> method for
   * further discussion on each entry in the dictionary.
   *
   * @see #loadDictionary
   */
   public StrokeDef[] activeDictionary;

   private static StrokeDef[] custom1;
   private static StrokeDef[] custom2;
   private static StrokeDef[] custom3;
   private static StrokeDef[] custom4;
   private static StrokeDef[] custom5;
   private static final int POS = 1;
   private static final int NEG = 0;

   private String undefinedStroke;
   private boolean flipX = false;
   private boolean flipY = false;
   private int tapThreshhold;
   private float aspectRatio;
   private final int DC = 4; // don't care
   private int quadFirst  = 0;
   private int quadSecond  = 0;
   private int quadPenultimate = 0;
   private int quadLast  = 0;
   private float kx = 0.0f;
   private float ky = 0.0f;
   private int startx = 0;
   private int starty = 0;
   private int stopx = 0;
   private int stopy = 0;

   /** Construct a Unistroke object.<p>
   *
   * Once a Unistroke object is declared, character recognition
   * may proceed immediately via the <code>recognize()</code>
   * method.
   *
   * The default dictionary is the graffiti dictionary.<p>
   */
   Unistroke()
   {
      tapThreshhold = TAP_THRESHHOLD;
      aspectRatio = ASPECT_RATIO;
      undefinedStroke = UNRECOGNIZED_STROKE;
      setDictionary(GRAFFITI); // set default dictionary to graffiti
   }

   /** Change the threshold setting for a tap.<p>
   *
   * The definition of a stylus or finger tap is any stroke with less than a certain
   * threshold of samples.  The default threshold is 5.<p>
   *
   * The <code>recognize()</code> method returns "=TAP" for any stroke
   * meeting this criterion.<p>
   *
   * @param n an integer representing the number of samples below which
   * a stroke is considered a tap.
   */
   public void setTapThreshhold(int n)
   {
      tapThreshhold = n;
   }

   /** Change the return string for an undefined stroke.<p>
   *
   * By default, the recognizer returns the string "<code>#</code>"
   * (hash symbol) for unrecognized strokes.<p>
   *
   * @param s the new return string for unrecognized strokes
   */
   public void setUndefined(String undefinedArg)
   {
      undefinedStroke = undefinedArg;
   }

   /** Change the aspect ratio for horizontal or vertical strokes.
   *
   * The default aspect ratio is 0.2.<p>
   *
   * @param d a <code>double</code> representing the new aspect ratio.<p>
   * </dl><p>
   *
   * <hr><p>
   * Discussion:<p>
   *
   * Horizontal or vertical strokes pose a special problem for the
   * recognizer because the stroke shape is grossly distorted when the
   * sample points are normalized.  To accommodate this, the recognition
   * of horizontal or vertical strokes is done before normalizing the stroke points.
   * <p>
   *
   * A stroke is recognized as horizontal or vertical if the aspect
   * ratio &mdash; the thickness-to-length ratio &mdash;
   * of the unnormalized sample points is less than a certain
   * threshold.  The default is 0.2.<p>
   */
   public void setAspectRatio(float d)
   {
      aspectRatio = d;
   }

   /** Flip the orientation of the <i>y</i> coordinates.<p>
   *
   * The sample points passed to the <code>recognize()</code>
   * method are assumed to be consistent with Java's
   * AWT coordinate system: namely, <i>x</i> coordinates increase moving
   * to the right across the display, and <i>y</i> coordinates increase moving
   * down the display.  The <code>flipY()</code> method
   * can be used to reverse
   * the orientation of the <i>y</i> coordinates, if necessary.<p>
   *
   * This method needs to be called once only, after instantiating
   * the <code>Unistroke</code> object.  If the method is called
   * a second time, the original orientation is restored.
   */
   public void flipY()
   {
      flipY = !flipY;
   }

   /** Flip the orientation of the <i>x</i> coordinates.
   *
   * See the <code>flipY()</code> method for discussion.
   */
   public void flipX()
   {
      flipX = !flipX;
   }

   /** Set or change the active dictionary. For example,<p>
   *
   * <pre>
   *     setDictionary(Unistroke.DIGITS);
   * </pre>
   *
   * changes the currently active dictionary to the digits
   * dictionary.<p>
   *
   * @param d an integer representing a dictionary.<p>
   * The following
   * dictionaries are defined:<p>
   *
   * <ul>
   * <li>   
   * <code>Unistroke.GRAFFITI</code> - built-in Graffiti dictionary (default)
   * <li>
   * <code>Unistroke.UNISTROKES</code> - built-in Unistrokes dictionary
   * <li>
   * <code>Unistroke.DIGITS</code> - build-in digits dictionary
   * <li>
   * <code>Unistroke.CUSTOM1</code> - user-provided custom dictionary #1
   * <li>
   * <code>Unistroke.CUSTOM2</code> - user-provided custom dictionary #2
   * <li>
   * <code>Unistroke.CUSTOM3</code> - user-provided custom dictionary #3 
   * <li>
   * <code>Unistroke.CUSTOM4</code> - user-provided custom dictionary #4
   * <li>
   * <code>Unistroke.CUSTOM5</code> - user-provided custom dictionary #5
   * </ul>
   */
   public void setDictionary(int d)
   {
	  currentDictionary = d;
      if (d == GRAFFITI)
         activeDictionary = gsd;      // graffiti stroke dictionary
      else if (d == UNISTROKES)
         activeDictionary = usd;      // unistrokes stroke dictionary
      else if (d == DIGITS)
         activeDictionary = dsd;      // digits stroke dictionary
      else if (d == CUSTOM1)
         activeDictionary = custom1;  // custom1 stroke dictionary
      else if (d == CUSTOM2)
         activeDictionary = custom2;  // custom2 stroke dictionary
      else if (d == CUSTOM3)
         activeDictionary = custom3;  // custom3 stroke dictionary
      else if (d == CUSTOM4)
         activeDictionary = custom4;  // custom4 stroke dictionary
      else if (d == CUSTOM5)
         activeDictionary = custom5;  // custom5 stroke dictionary
   }

   // flip the orientation of the sample points, as per flipX & flipY
   private void flip(int[] x, int[] y, int n)
   {
      for (int i = 0; i < n; ++i)
      {
         if (flipX)
            x[i] = Integer.MAX_VALUE - x[i];
         if (flipY)
            y[i] = Integer.MAX_VALUE - y[i];
      }
   }

   /** Perform handwriting recognition on gestures. This is the basic
   * method to convert a series of <i>x-y</i> sample points to a character.
   * The returned string is one of the following:<p>
   *
   * <ul>
   * <li>A string containing the entry in the
   * dictionary corresponding to the stroke, if recognition was
   * successful
   * <li>The string "<code>#</code>" (hash symbol),
   * if a match for the stroke was not found
   * <li>The string "<code>=TAP</code>" if the stroke meets the criterion for
   * a tap (i.e., it contains less than a certain threshold of samples)
   * </ul>
   *
   * @param x an integer array of <i>x</i> sample points.
   * @param y an integer array of <i>y</i> sample points.
   * @return a string representation of the recognized stroke<p>
   */
   public String recognize(int[] x, int[] y)
   {
      // make copies of arrays, so original data are intact
      int[] xx = new int[x.length];
      int[] yy = new int[x.length];
      for (int i = 0; i < x.length; ++i)
      {
         xx[i] = x[i];
         yy[i] = y[i];
      }

      // flip orientation of sample points, if necessary
      flip(xx, yy, xx.length);

      // do the recognition and return result
      return toCharacter(xx, yy, xx.length);
   }

   /** Perform handwriting recognition (3-arg version).
   * @param x an integer array of <i>x</i> sample points.
   * @param y an integer array of <i>y</i> sample points.
   * @param length an integer representing the number of elements
   * in the arrays to process
   * @return a string representation of the recognized stroke<p>
   */
   public String recognize(int[] x, int[] y, int length)
   {
      // make copies of arrays, so original data are intact
      int[] xx = new int[length];
      int[] yy = new int[length];
      for (int i = 0; i < length; ++i)
      {
         xx[i] = x[i];
         yy[i] = y[i];
      }

      // flip orientation of sample points, if necessary
      flip(xx, yy, length);

      // do the recognition and return result
      return toCharacter(xx, yy, length);
   }

   /** Perform handwriting recognition (<code>Point</code> class version).
   * @param p an array of <code>Point</code> objects.
   * @param length an integer representing the number of elements
   * in the <code>Point</code> array to process
   * @return a string representation of the recognized stroke<p>
   */
   public String recognize(Point p[], int length)
   {
      // make arrays
      int[] xx = new int[length];
      int[] yy = new int[length];
      for (int i = 0; i < length; ++i)
      {
         xx[i] = (int)p[i].x;
         yy[i] = (int)p[i].y;
      }

      // flip orientation of sample points, if necessary
      flip(xx, yy, length);

      // do the recognition and return result
      return toCharacter(xx, yy, length);
   }

   /** Calculate and return the features of a unistroke (3-arg version) */
   public String getFeatures(int[] x, int[] y, int length)
   {
      // perform dummy recognition (mostly to compute the features)
      String s = recognize(x, y, length);

      if (s.equals("=TAP"))
         return s;
      else
         return quadFirst + ", " + quadSecond + ", " + quadPenultimate + ", " + quadLast + ", " +
         	trim(kx, 3) + ", " + trim(ky, 3) + ", " +
             startx + ", " + starty + ", " + stopx + ", " + stopy;
   }

   /** Calculate and return the features of a unistroke.
   * The returned string is one of the following:<p>
   *
   * <ul>
   * <li>A string containing a set of ten comma-delimited
   * features for the passed unistroke
   * <li>The string "<code>=TAP</code>" for a stylus or finger tap
   * </ul><p>
   *
   * For example, the following
   * could be the returned string for a unistroke resembling a
   * Graffiti "a":<p>
   *
   * <pre>
   *     2, 3, 0, 1, 1.010, 1.857, 1, 1, 1, 0
   * </pre>
   *
   * The ten values above represent the following unistroke features:<p>
   *
   * <pre>
   *     quadFirst        = 2
   *     quadSecond       = 3
   *     quadsPenultimate = 0
   *     quadLast         = 1
   *     kx               = 1.010
   *     ky               = 1.855
   *     startx           = 1
   *     starty           = 1
   *     stopx            = 1
   *     stopy            = 0
   * </pre>
   *
   * @param x an integer array of <i>x</i> sample points.
   * @param y an integer array of <i>y</i> sample points.
   * @return a string containing the features of the passed unistroke,
   * or "<code>=TAP</code>" for a unistroke representing a
   *  tap.
   *
   */
   public String getFeatures(int[] x, int[] y)
   {
      // perform dummy recognition (mostly to compute the features)
      String s = recognize(x, y);

      if (s.equals("=TAP"))
         return s;
      else
         return quadFirst + ", " + quadSecond + ", " + quadPenultimate + ", " + quadLast + ", " +
         	trim(kx, 3) + ", " + trim(ky, 3) + ", " + 
            startx + ", " + starty + ", " + stopx + ", " + stopy;
   }

   //========================================================
   // All the work is done here!
   //========================================================
   private String toCharacter(int[] x, int[] y, int n)
   {
      StrokeDef[] d = activeDictionary; // just make a copy for convenience

      float xMin = Float.MAX_VALUE;
      float xMax = Float.MIN_VALUE;
      float yMin = Float.MAX_VALUE;
      float yMax = Float.MIN_VALUE;

      float[] nx = new float[n];
      float[] ny = new float[n];
      kx = 0.0f;
      ky = 0.0f;
      float kz = 0.0f; // cumulative distance of the entire stroke
      startx = 0;
      starty = 0;
      stopx = 0;
      stopy = 0;
 
      // if less than "tapThreshhold" points, it's a "tap"
      if (n < tapThreshhold)
      {
    	  //Log.i("MYDEBUG", "TAP! n=" + n);
    	  return "=TAP";
      }

      //--------------------------
      // Calculate Stroke Features
      //--------------------------

      // Find min and max for sample points on x and y axes 
      for (int j = 0; j < n; ++j)
      {
         if (x[j] < xMin) xMin = x[j];  // x lower 
         if (x[j] > xMax) xMax = x[j];  // x upper 
         if (y[j] < yMin) yMin = y[j];  // y lower 
         if (y[j] > yMax) yMax = y[j];  // y upper 
      }

      // Normalize points to fit in a "unit" bounding box 
      for (int j = 0; j < n; ++j)
      {
         if ((xMax - xMin) != 0.0)
            nx[j] = x[j] / (xMax - xMin) - xMin / (xMax - xMin);
         else
            nx[j] = 0.5f;

         if ((yMax - yMin) != 0.0)
            ny[j] = y[j] / (yMax - yMin) - yMin / (yMax - yMin);
         else
            ny[j] = 0.5f;
      }

      //---------------------------------------------------------
      // Calculate cumulative length of stroke along x and y axes
      //----------------------------------------------------------
      for (int j = 1; j < n; ++j)
      {
         kx += Math.abs(nx[j] - nx[j - 1]);
         ky += Math.abs(ny[j] - ny[j - 1]);
         kz += (float)Math.sqrt((nx[j] - nx[j - 1]) * (nx[j] - nx[j - 1]) + (ny[j] - ny[j - 1]) * (ny[j] - ny[j - 1]));
      }

      //-----------------------------------------------
      // Determine 1st, 2nd, 2nd last, & last quadrants
      //-----------------------------------------------
      // where...           -------
      // 0 = upper right   | 3 | 0 |
      // 1 = lower right   |-------|
      // 2 = lower left    | 2 | 1 |
      // 3 = upper left     -------
      //
      quadFirst = findQuad(nx[0], ny[0]);
      quadLast  = findQuad(nx[n - 1], ny[n - 1]);
      int j;
      
      quadSecond = quadFirst;      
      for (j = 0; j < n && findQuad(nx[j], ny[j]) == quadFirst; ++j)
    	  ;
      quadSecond = findQuad(nx[j], ny[j]);

      quadPenultimate = quadLast;
      for (j = n - 1; j >= 0 && findQuad(nx[j], ny[j]) == quadLast; --j)
    	  ;
      quadPenultimate = findQuad(nx[j], ny[j]);        

      //-------------------------------------------------------
      // Determine starting and terminating direction of stroke
      //-------------------------------------------------------
      // Note: Starting direction is based on a comparison between the first sample point and the sample
      // point corresponding to approximately 10% of the cumulative distance along the stroke. 
      // The stopping direction is based on a comparison between the last sample point and the sample
      // point corresponding to about 90% of the cumulative distance along the stroke.
      // E.g., if startx = 1, movement is to the right
      //       if starty = 1, movement is down
      
      int idxNearStart = -1;
      int idxNearEnd = -1;
      float zTemp = 0; // cumulative distance along the stroke
      for (j = 1; j < n; ++j)
      {
    	  zTemp += (float)Math.sqrt((nx[j] - nx[j - 1]) * (nx[j] - nx[j - 1]) + (ny[j] - ny[j - 1]) * (ny[j] - ny[j - 1]));
    	  if (idxNearStart == -1 && zTemp >= (kz * 0.10))
    		  idxNearStart = j;
    	  if (idxNearEnd == -1 && zTemp >= (kz * 0.90))
    	  {
    		  idxNearEnd = j;
    		  break;
    	  }
      }
      
      startx = nx[idxNearStart] - nx[0] > 0 ? POS : NEG; // POS = right, NEG = left
      starty = ny[idxNearStart] - ny[0] < 0 ? POS : NEG; // POS = down, NEG = up 
      stopx = nx[n - 1] - nx[idxNearEnd] > 0 ? POS : NEG;
      stopy = ny[n - 1] - ny[idxNearEnd] < 0 ? POS : NEG;
     
      // -------------------------------------
      // Check for straight-line strokes first
      // -------------------------------------

      String stroke = "";    
      
      //----------------------------------------------------------------
      // Vertical or horizontal strokes (use unnormalized sample points)
      //----------------------------------------------------------------
      if ((xMax - xMin) < aspectRatio * (yMax - yMin) && ny[0] > ny[n - 1]) // north
    	  stroke = "=N";  
      else if ((yMax - yMin) < aspectRatio * (xMax - xMin) && nx[0] < nx[n - 1]) // east
    	  stroke = "=E";
      else if ((yMax - yMin) < aspectRatio * (xMax - xMin) && nx[0] > nx[n - 1]) // west
    	  stroke = "=W";
      else if ((xMax - xMin) < aspectRatio * (yMax - yMin) && ny[0] < ny[n - 1]) // south
      {
    	  if (currentDictionary == GRAFFITI)
    		  stroke = "i"; 
    	  else if (currentDictionary == DIGITS)
    		  stroke = "1";
    	  else
    		  stroke = "=S";
      }
      
      if (stroke.length() == 0) // nothing yet...
      {
	      // -------------------------------------------------------------     
	      // Diagonal straight-line strokes (use normalized sample points)
	      // -------------------------------------------------------------
	      double adjustedAspectRatio = 1.4 * aspectRatio;  // to accommodate the diagonal (it's a trig thing)
	      
	      boolean equalDiagonal = true;    // x and y both increasing or both decreasing along array
	      boolean oppositeDiagonal = true; // x and y moving in opposite directions along stroke
	      for (int i = 0; i < n; ++i)
	      {
	    	  if (nx[i] < ny[i] - adjustedAspectRatio || nx[i] > ny[i] + adjustedAspectRatio)
	    		  equalDiagonal = false;
	    	  if (1 - nx[i] < ny[i] - adjustedAspectRatio || 1 - nx[i] > ny[i] + adjustedAspectRatio)
	    		  oppositeDiagonal = false;
	    	  if (!equalDiagonal && !oppositeDiagonal)
	    		  break; // no need to check any further
	      }	
	      
	      if (equalDiagonal && nx[0] < nx[n - 1])  // x is decreasing
	    	  stroke = "=SE";
	      else if (oppositeDiagonal && nx[0] < nx[n - 1]) // x is increasing
	    	  stroke = "=NE";
	      else if (oppositeDiagonal && nx[0] > nx[n - 1]) // x is decreasing
	    	  stroke = "=SW";
	      else if (equalDiagonal && nx[0] > nx[n - 1]) // x is increasing
	    	  stroke = "=NW"; 
      }
      
      if (stroke.length() == 0) // not a straight-line stroke
      {      
	      //-----------------------------------
	      // Look for a match in the dictionary
	      //-----------------------------------
	      dictionaryMatches = new Vector<String>();
	      
	      // NOTE: This routine finds the "first match" rather than the "best match".  Finding the best match makes more sense,
	      // so this is something for the "to do" list.
	      
	      for (j = 0; j < d.length; j++)
	      {
	    	  if ( (quadFirst  == d[j].getDEFquadf()  || d[j].getDEFquadf()  == DC) &&
	               (quadSecond  == d[j].getDEFquads()  || d[j].getDEFquads()  == DC) &&
	               (quadPenultimate == d[j].getDEFquadsl() || d[j].getDEFquadsl() == DC) &&
	               (quadLast  == d[j].getDEFquadl()  || d[j].getDEFquadl()  == DC) &&
	               kx >= d[j].getDEFkxmin() &&
	               kx <= d[j].getDEFkxmax() &&
	               ky >= d[j].getDEFkymin() &&
	               ky <= d[j].getDEFkymax() &&
	               (startx == d[j].getDEFstartX() || d[j].getDEFstartX() == DC) &&
	               (starty == d[j].getDEFstartY() || d[j].getDEFstartY() == DC) &&
	               (stopx  == d[j].getDEFstopX()  || d[j].getDEFstopX()  == DC) &&
	               (stopy  == d[j].getDEFstopY()  || d[j].getDEFstopY()  == DC) )
				{
					dictionaryMatches.add(d[j].getDEFsymbol());				
				}
	      }	
	    	
	      // Take the first match (TO DO: find the "best" match)
		  stroke = dictionaryMatches.size() > 0 ? (String)dictionaryMatches.elementAt(0) : undefinedStroke;       
      }
      return stroke;
   }	

   private int findQuad(double x, double y)
   {
      if (x >= 0.5 && y <= 0.5) return 0; // upper right 
      if (x >= 0.5 && y >  0.5) return 1; // lower right 
      if (x <  0.5 && y >  0.5) return 2; // lower left  
      if (x <  0.5 && y <= 0.5) return 3; // upper left  
      return 0;
   }
   
	// trim and round a float to the specified number of decimal places
	private float trim(float f, int decimalPlaces)
	{
		return (int)(f * 10 * decimalPlaces + 0.5f) / (float)(10 * decimalPlaces);
	}
}